/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * r4xx0.c: R4000 processor variant specific MMU/Cache routines.
 *
 * Copyright (C) 1996 David S. Miller (dm@engr.sgi.com)
 * Copyright (C) 1997, 1998, 1999, 2000 Ralf Baechle ralf@gnu.org
 *
 * To do:
 *
 *  - this code is a overbloated pig
 *  - many of the bug workarounds are not efficient at all, but at
 *    least they are functional ...
 */
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/config.h>
#include <linux/module.h>

#include <asm/cpu.h>
#include <asm/bootinfo.h>
#include <asm/mmu_context.h>
#include <asm/pgtable.h>
#include <asm/system.h>

#undef DEBUG_TLB
#undef DEBUG_TLBUPDATE

extern char except_vec0_nevada, except_vec0_r4000, except_vec0_r4600;
extern char except_vec0_r4000_nop;

/* CP0 hazard avoidance. */
#define BARRIER __asm__ __volatile__(".set noreorder\n\t" \
				     "nop; nop; nop; nop; nop; nop;\n\t" \
				     ".set reorder\n\t")

void local_flush_tlb_all(void)
{
	unsigned long flags;
	unsigned long old_ctx;
	int entry;

#ifdef DEBUG_TLB
	printk("[tlball]");
#endif

	__save_and_cli(flags);
	/* Save old context and create impossible VPN2 value */
	old_ctx = (get_entryhi() & 0xff);
	set_entrylo0(0);
	set_entrylo1(0);
	BARRIER;

	entry = get_wired();

	/* Blast 'em all away. */
	while (entry < mips_cpu.tlbsize) {
		/*
		 * Make sure all entries differ.  If they're not different
		 * MIPS32 will take revenge ...
		 */
		set_entryhi(KSEG0 + entry*0x2000);
		set_index(entry);
		BARRIER;
		tlb_write_indexed();
		BARRIER;
		entry++;
	}
	BARRIER;
	set_entryhi(old_ctx);
	__restore_flags(flags);
}

void local_flush_tlb_mm(struct mm_struct *mm)
{
	if (mm->context != 0) {
		unsigned long flags;

#ifdef DEBUG_TLB
		printk("[tlbmm<%d>]", mm->context);
#endif
		__save_and_cli(flags);
		get_new_mmu_context(mm, smp_processor_id());
		if (mm == current->active_mm)
			set_entryhi(mm->context & 0xff);
		__restore_flags(flags);
	}
}

void local_flush_tlb_range(struct mm_struct *mm, unsigned long start,
				unsigned long end)
{
	if (mm->context != 0) {
		unsigned long flags;
		int size;
#ifdef CONFIG_BMIPS3300_WAR
		unsigned long prev_addr;
		pgd_t *pgdp;
		pmd_t *pmdp;
		pte_t *ptep;
		int flush_all = 0;
#endif

#ifdef DEBUG_TLB
		printk("[tlbrange<%02x,%08lx,%08lx>]", (mm->context & 0xff),
		       start, end);
#endif
		__save_and_cli(flags);
#ifdef CONFIG_BMIPS3300_WAR
		prev_addr = start - 0x2000;
		pgdp = pgd_offset(mm, prev_addr);
		pmdp = pmd_offset(pgdp, prev_addr);
		ptep = pte_offset(pmdp, prev_addr);
		if ((pte_val(ptep[1]) & (0x80000000 | _PAGE_VALID)) ==
		                        (0x80000000 | _PAGE_VALID)) {
			flush_all = 1;
		}
#endif
		size = (end - start + (PAGE_SIZE - 1)) >> PAGE_SHIFT;
		size = (size + 1) >> 1;
#ifdef CONFIG_BMIPS3300_WAR
		if (!flush_all && size <= mips_cpu.tlbsize/2) {
#else
		if (size <= mips_cpu.tlbsize/2) {
#endif
			int oldpid = (get_entryhi() & 0xff);
			int newpid = (mm->context & 0xff);

			start &= (PAGE_MASK << 1);
			end += ((PAGE_SIZE << 1) - 1);
			end &= (PAGE_MASK << 1);
			while (start < end) {
				int idx;

				set_entryhi(start | newpid);
				start += (PAGE_SIZE << 1);
				BARRIER;
				tlb_probe();
				BARRIER;
				idx = get_index();
				set_entrylo0(0);
				set_entrylo1(0);
				if (idx < 0)
					continue;
				/* Make sure all entries differ. */
				set_entryhi(KSEG0 + idx*0x2000);
				BARRIER;
				tlb_write_indexed();
				BARRIER;
			}
			set_entryhi(oldpid);
		} else {
			get_new_mmu_context(mm, smp_processor_id());
			if (mm == current->active_mm)
				set_entryhi(mm->context & 0xff);
		}
		__restore_flags(flags);
	}
}

void local_flush_tlb_page(struct vm_area_struct *vma, unsigned long page)
{
	if (!vma || vma->vm_mm->context != 0) {
		unsigned long flags;
		int oldpid, newpid, idx;
#ifdef CONFIG_BMIPS3300_WAR
		unsigned long prev_addr;
		pgd_t *pgdp;
		pmd_t *pmdp;
		pte_t *ptep;
#endif

#ifdef DEBUG_TLB
		printk("[tlbpage<%d,%08lx>]", vma->vm_mm->context, page);
#endif
		newpid = (vma->vm_mm->context & 0xff);
		page &= (PAGE_MASK << 1);
		__save_and_cli(flags);
#ifdef CONFIG_BMIPS3300_WAR
		prev_addr = page - 0x2000;
		pgdp = pgd_offset(vma->vm_mm, prev_addr);
		pmdp = pmd_offset(pgdp, prev_addr);
		ptep = pte_offset(pmdp, prev_addr);
		if ((pte_val(ptep[1]) & (0x80000000 | _PAGE_VALID)) !=
		                        (0x80000000 | _PAGE_VALID)) {
#endif
		oldpid = (get_entryhi() & 0xff);
		set_entryhi(page | newpid);
		BARRIER;
		tlb_probe();
		BARRIER;
		idx = get_index();
		set_entrylo0(0);
		set_entrylo1(0);
		if(idx < 0)
			goto finish;
		/* Make sure all entries differ. */
		set_entryhi(KSEG0+idx*0x2000);
		BARRIER;
		tlb_write_indexed();

	finish:
		BARRIER;
		set_entryhi(oldpid);
#ifdef CONFIG_BMIPS3300_WAR
		} else {
			struct mm_struct *mm = vma->vm_mm;

			get_new_mmu_context(mm, smp_processor_id());
			if (mm == current->active_mm)
				set_entryhi(mm->context & 0xff);
		}
#endif
		__restore_flags(flags);
	}
}

#ifdef CONFIG_BMIPS3300_WAR
static inline int check_page_inst(unsigned long page_addr)
{
	u32* page_inst = (u32*)page_addr;

	return (((page_inst[0x0ff0/4] & 0xa0000000) == 0xa0000000) &&
	        ((page_inst[0x0ff4/4] & 0xa0000000) == 0x80000000) &&
	        ((page_inst[0x0ff8/4] & 0xa0000000) == 0x80000000));
}
#endif

/* We will need multiple versions of update_mmu_cache(), one that just
 * updates the TLB with the new pte(s), and another which also checks
 * for the R4k "end of page" hardware bug and does the needy.
 */
void update_mmu_cache(struct vm_area_struct * vma, unsigned long address,
		      pte_t pte)
{
	unsigned long flags;
	pgd_t *pgdp;
	pmd_t *pmdp;
	pte_t *ptep;
	int idx, pid;

	/*
	 * Handle debugger faulting in for debugee.
	 */
	if (current->active_mm != vma->vm_mm)
		return;

	pid = get_entryhi() & 0xff;

#ifdef DEBUG_TLB
	if((pid != (vma->vm_mm->context & 0xff)) || (vma->vm_mm->context == 0)) {
		printk("update_mmu_cache: Wheee, bogus tlbpid mmpid=%d tlbpid=%d\n",
		       (int) (vma->vm_mm->context & 0xff), pid);
	}
#endif

	__save_and_cli(flags);
	address &= (PAGE_MASK << 1);
	set_entryhi(address | (pid));
	pgdp = pgd_offset(vma->vm_mm, address);
	BARRIER;
	tlb_probe();
	BARRIER;
	pmdp = pmd_offset(pgdp, address);
	idx = get_index();
	ptep = pte_offset(pmdp, address);
	BARRIER;
#ifdef CONFIG_BMIPS3300_WAR
	if (vma->vm_flags & VM_EXEC) {
		if (pte_present(ptep[0]) && !(pte_val(ptep[0]) & 0x40000000)) {
			pte_val(ptep[0]) |= _PAGE_VALID | _PAGE_ACCESSED;
			pte_val(ptep[0]) &= 0x7fffffff;
		}
		if (pte_present(ptep[1]) && !(pte_val(ptep[1]) & 0x40000000)) {
			pte_val(ptep[1]) |= _PAGE_VALID | _PAGE_ACCESSED;
			pte_val(ptep[1]) &= 0x7fffffff;
		}
	}
	set_entrylo0(pte_val(ptep[0]) >> 6);
	set_entrylo1(pte_val(ptep[1]) >> 6);
#else
	set_entrylo0(pte_val(*ptep++) >> 6);
	set_entrylo1(pte_val(*ptep) >> 6);
#endif
	set_entryhi(address | (pid));
	BARRIER;
	if (idx < 0) {
		tlb_write_random();
	} else {
		tlb_write_indexed();
	}
	BARRIER;
#ifdef CONFIG_BMIPS3300_WAR
	if (vma->vm_flags & VM_EXEC) {
		int need_flush = 0;

		if (pte_present(ptep[0]) && !(pte_val(ptep[0]) & 0x40000000)) {
			if (check_page_inst(address)) {
				struct vm_area_struct* vma0;
				struct vm_area_struct* vma1;

				/* Only mark the page if both it and the following
				 * page are executable */
				vma0 = find_vma(vma->vm_mm, address + 0x0000);
				vma1 = find_vma(vma->vm_mm, address + 0x1000);
				if (vma0 && (vma0->vm_flags & VM_EXEC) && !(vma0->vm_flags & VM_WRITE) &&
				    vma1 && (vma1->vm_flags & VM_EXEC)) {
					pte_val(ptep[0]) |= 0x80000000;
					if (!pte_present(ptep[1])) {
						need_flush = 1;
					}
				}
			}
		}

		if (pte_present(ptep[1]) && !(pte_val(ptep[1]) & 0x40000000)) {
			if (check_page_inst(address + 0x1000)) {
				struct vm_area_struct* vma1;
				struct vm_area_struct* vma2;

				/* Only mark the page if both it and the following
				 * page are executable */
				vma1 = find_vma(vma->vm_mm, address + 0x1000);
				vma2 = find_vma(vma->vm_mm, address + 0x2000);
				if (vma1 && (vma1->vm_flags & VM_EXEC) && !(vma1->vm_flags & VM_WRITE) &&
				    vma2 && (vma2->vm_flags & VM_EXEC)) {
					pte_val(ptep[1]) |= 0x80000000;
					need_flush = 1;
				}
			}
		}

		if (need_flush) {
			local_flush_tlb_page(vma, address);
		}
	}
#endif
	set_entryhi(pid);
	BARRIER;
	__restore_flags(flags);
}

#ifdef CONFIG_BMIPS3300_WAR
extern asmlinkage void
do_page_fault(struct pt_regs *regs, unsigned long write, unsigned long address);

#define BARRIER4 __asm__ __volatile__(".set noreorder\n" \
				      "nop; nop; nop; nop;\n" \
				      ".set reorder\n")

asmlinkage void do_fixup_multiple(struct pt_regs *regs, unsigned long address)
{
	struct mm_struct *mm;
	unsigned long flags;
	int ctx, i, second_pass;

restart:
	__save_and_cli(flags);

	mm = current->active_mm;
	ctx = mm->context & 0xff;
	second_pass = 0;

	while (second_pass < 2) {
		unsigned long pcurr, pdcurr;
		int need_valid;
		pgd_t *pgdp;
		pmd_t *pmdp;
		pte_t *ptep;

		pdcurr = (address & PGDIR_MASK) + PGDIR_SIZE;
		pcurr = address & (PAGE_MASK << 1);

		pgdp = pgd_offset(mm, pcurr);
		pmdp = pmd_offset(pgdp, pcurr);
		ptep = pte_offset(pmdp, pcurr);

		need_valid = 0;

		for (i = 0; i < mips_cpu.tlbsize; i++) {
			if (pcurr == (address & PAGE_MASK)) {
				need_valid = 1;
			}

			if (!second_pass) {
				/*
				 * First pass: fault in pages that need it
				 */
				if (need_valid && !pte_present(ptep[0])) {
					__restore_flags(flags);

					do_page_fault(regs, 0, pcurr);
					goto restart;
				}
			} else {
				/*
				 * Second pass: make the pages valid, invalidate old
				 * TLB entry if mapped, and write new TLB entry for
				 * each page pair
				 */
				if (need_valid && !(pte_val(ptep[0]) & _PAGE_VALID)) {
					pte_val(ptep[0]) |= _PAGE_VALID | _PAGE_ACCESSED;
				}

				if (pcurr & PAGE_SIZE) {
					int idx;

					set_entryhi((pcurr ^ PAGE_SIZE) | ctx);
					BARRIER4;
					tlb_probe();
					BARRIER4;
					idx = get_index();
					if (idx >= 0) {
						set_entryhi(KSEG0 + idx*0x2000);
						set_entrylo0(0);
						set_entrylo1(0);
						BARRIER4;
						tlb_write_indexed();
						BARRIER4;
						set_entryhi((pcurr ^ PAGE_SIZE) | ctx);
					}
					set_entrylo0((pte_val(ptep[-1]) << 1) >> 7);
					set_entrylo1((pte_val(ptep[ 0]) << 1) >> 7);
					BARRIER4;
					tlb_write_random();
				}
			}

			need_valid = ((pte_val(ptep[0]) & 0xc0000080) == 0x80000080);
			if (!need_valid && (pcurr & PAGE_SIZE)) {
				break;
			}

			pcurr += PAGE_SIZE;
			if (pcurr > pdcurr) {
				pmdp++;
				if (pmd_none(*pmdp)) {
					break;
				}
				pdcurr += PGDIR_SIZE;
				ptep = pte_offset(pmdp, pcurr);
			} else {
				ptep++;
			}
		}

		second_pass++;
	}

	__restore_flags(flags);

	if (i == mips_cpu.tlbsize) {
		static int warned = 0;

		if (!warned) {
			printk(KERN_WARNING "BMIPS3300 WAR: Overflow at %08x\n",
			       (unsigned int)address);
			warned = 1;
		}
	}
}

/* This is not especially optimal, but it doesn't really need to be */
void update_mmu_module(struct module *mod)
{
	unsigned long addr, size, end;
	int i;

	addr = 0;
	size = 0;

	for (i = 0; i < mod->nsyms; i++) {
		const char* name = mod->syms[i].name;

		if (strncmp(name, "__insmod", 8) == 0) {
			const char* text_pos = strstr(name, "_S.text_L");

			if (text_pos) {
				unsigned long max_addr;

				addr = mod->syms[i].value;
				size = simple_strtoul(text_pos+9, NULL, 10);

				max_addr = (unsigned long)mod + mod->size;
				if ((addr < (unsigned long)mod) ||
				    (addr + size > max_addr)) {
					addr = 0;
				}

				break;
			}
		}
	}

	if (addr == 0) {
		printk(KERN_WARNING "BMIPS3300 WAR: Couldn't locate text "
		       "section in module %s\n", mod->name);
		addr = (unsigned long)mod;
		size = mod->size;
	}

	end = (addr + size - 1) & PAGE_MASK;
	addr &= PAGE_MASK;

	while (addr < end) {
		if (check_page_inst(addr)) {
			pgd_t *pgdp;
			pmd_t *pmdp;
			pte_t *ptep;

			pgdp = pgd_offset(&init_mm, addr);
			pmdp = pmd_offset(pgdp, addr);
			ptep = pte_offset(pmdp, addr);

			pte_val(ptep[0]) |= 0x80000000;
		}

		addr += PAGE_SIZE;
	}

	local_flush_tlb_all();
}
#endif

#if 0
static void r4k_update_mmu_cache_hwbug(struct vm_area_struct * vma,
				       unsigned long address, pte_t pte)
{
	unsigned long flags;
	pgd_t *pgdp;
	pmd_t *pmdp;
	pte_t *ptep;
	int idx;

	__save_and_cli(flags);
	address &= (PAGE_MASK << 1);
	set_entryhi(address | (get_entryhi() & 0xff));
	pgdp = pgd_offset(vma->vm_mm, address);
	tlb_probe();
	pmdp = pmd_offset(pgdp, address);
	idx = get_index();
	ptep = pte_offset(pmdp, address);
	set_entrylo0(pte_val(*ptep++) >> 6);
	set_entrylo1(pte_val(*ptep) >> 6);
	BARRIER;
	if (idx < 0)
		tlb_write_random();
	else
		tlb_write_indexed();
	BARRIER;
	__restore_flags(flags);
}
#endif

void __init add_wired_entry(unsigned long entrylo0, unsigned long entrylo1,
		     unsigned long entryhi, unsigned long pagemask)
{
	unsigned long flags;
	unsigned long wired;
	unsigned long old_pagemask;
	unsigned long old_ctx;

	__save_and_cli(flags);
	/* Save old context and create impossible VPN2 value */
	old_ctx = get_entryhi() & 0xff;
	old_pagemask = get_pagemask();
	wired = get_wired();
	set_wired(wired + 1);
	set_index(wired);
	BARRIER;
	set_pagemask(pagemask);
	set_entryhi(entryhi);
	set_entrylo0(entrylo0);
	set_entrylo1(entrylo1);
	BARRIER;
	tlb_write_indexed();
	BARRIER;

	set_entryhi(old_ctx);
	BARRIER;
	set_pagemask(old_pagemask);
	local_flush_tlb_all();
	__restore_flags(flags);
}

/*
 * Used for loading TLB entries before trap_init() has started, when we
 * don't actually want to add a wired entry which remains throughout the
 * lifetime of the system
 */

static int temp_tlb_entry __initdata;

__init int add_temporary_entry(unsigned long entrylo0, unsigned long entrylo1,
			       unsigned long entryhi, unsigned long pagemask)
{
	int ret = 0;
	unsigned long flags;
	unsigned long wired;
	unsigned long old_pagemask;
	unsigned long old_ctx;

	__save_and_cli(flags);
	/* Save old context and create impossible VPN2 value */
	old_ctx = get_entryhi() & 0xff;
	old_pagemask = get_pagemask();
	wired = get_wired();
	if (--temp_tlb_entry < wired) {
		printk(KERN_WARNING "No TLB space left for add_temporary_entry\n");
		ret = -ENOSPC;
		goto out;
	}

	set_index(temp_tlb_entry);
	BARRIER;
	set_pagemask(pagemask);
	set_entryhi(entryhi);
	set_entrylo0(entrylo0);
	set_entrylo1(entrylo1);
	BARRIER;
	tlb_write_indexed();
	BARRIER;

	set_entryhi(old_ctx);
	BARRIER;
	set_pagemask(old_pagemask);
out:
	__restore_flags(flags);
	return ret;
}

static void __init probe_tlb(unsigned long config)
{
	unsigned int prid, config1;

	prid = read_32bit_cp0_register(CP0_PRID) & 0xff00;
	if (prid == PRID_IMP_RM7000 || !(config & (1 << 31)))
		/*
		 * Not a MIPS32 complianant CPU.  Config 1 register not
		 * supported, we assume R4k style.  Cpu probing already figured
		 * out the number of tlb entries.
		 */
		return;

	config1 = read_mips32_cp0_config1();
	if (!((config >> 7) & 3))
		panic("No MMU present");
	else
		mips_cpu.tlbsize = ((config1 >> 25) & 0x3f) + 1;
}

void __init r4k_tlb_init(void)
{
	u32 config = read_32bit_cp0_register(CP0_CONFIG);

	/*
	 * You should never change this register:
	 *   - On R4600 1.7 the tlbp never hits for pages smaller than
	 *     the value in the c0_pagemask register.
	 *   - The entire mm handling assumes the c0_pagemask register to
	 *     be set for 4kb pages.
	 */
	probe_tlb(config);
	set_pagemask(PM_4K);
	write_32bit_cp0_register(CP0_WIRED, 0);
	temp_tlb_entry = mips_cpu.tlbsize - 1;
	local_flush_tlb_all();

	if ((mips_cpu.options & MIPS_CPU_4KEX)
	    && (mips_cpu.options & MIPS_CPU_4KTLB)) {
		if (mips_cpu.cputype == CPU_NEVADA)
			memcpy((void *)KSEG0, &except_vec0_nevada, 0x80);
		else if (mips_cpu.cputype == CPU_R4600)
			memcpy((void *)KSEG0, &except_vec0_r4600, 0x80);
		else
			memcpy((void *)KSEG0, &except_vec0_r4000, 0x80);
		flush_icache_range(KSEG0, KSEG0 + 0x80);
	}
}
