/*
 * gen-ioctl: Ioctl interface for generic ioctl calls
 * 
 * Copyright (C) 2005 TiVo Inc. All Rights Reserved.
 */

#include <linux/init.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/tivo-genioctl.h>
#include <linux/tivo-crashallassert.h>
#include <linux/sgat-dio.h>
#include <asm/scatterlist.h>
#include <linux/uio.h>

#ifdef CONFIG_TIVO_EIGER
#include <asm/bcm7038.h>
#endif

#define SKIP_SATA_PM_FW_UPGD 1

int (*upgd_fw_init)(void) = NULL;
int (*is_fw_upgd_reqd)(void) = NULL;

#ifdef CONFIG_KSTACK_GUARD
extern void test_kstack_guard_check(struct task_struct *p);
#endif

#define GIOCTL_MAJOR 243

int gioctl_open (struct inode *inode, struct file *filp){
        return 0;
}

int gioctl_ioctl (struct inode *inode, struct file *file, 
                unsigned int cmd, unsigned long arg) {

        if ( !arg ){
                printk(KERN_ERR "[gen_ioctl] Invalid arg\n");
                return -EINVAL;
        }
        if(_IOC_TYPE(cmd) != GIOCTL_IOC_MAGIC)
        {
                printk(KERN_ERR "[gen_ioctl] cmd (0x%x) mismatch, "
                                "called with 0x%x while expecting 0x%x\n", 
                                cmd, _IOC_TYPE(cmd), GIOCTL_IOC_MAGIC);
                return -EINVAL;
        }

        switch(cmd){
            case CRASHALLASSRT:
                return tivocrashallassert(arg);
            break;
#ifdef CONFIG_TICK_KILLER
            case TICK_KILLER:
                return tick_killer_start(arg);
            break;
#endif
            case UPDATE_EXEC_CTX:
                /*Ineffective ioctl. Handling for the backward compatibility purposes.
                 * Has be removed.*/
            break;
#ifdef CONFIG_KSTACK_GUARD
            case TEST_KSTACK_GUARD:
                test_kstack_guard_check(current);
                return 0;
            break;
#endif
            case TEST_SGAT_DIO:
            {
                struct sgat_dio_fops gfops;
                sgat_dio_req dio_req;
                int c;
                int k = 'z';

                sgat_dio_get_default_fops(&gfops);
                
                if ( sgat_dio_init_req(&gfops,&dio_req,(sgat_dio_hdr_t*)arg)){
                        printk("Error: sgat_dio_init_req\n");
                        return -EINVAL;
                }

                if ( sgat_dio_build_sg_list(&gfops, &dio_req, 1)){
                        printk("Error: sgat_dio_build_sg_list\n");
                        return -EINVAL;
                }

                struct  scatterlist *p = (struct  scatterlist *) dio_req.sclist;
                for ( c =0; c < dio_req.k_useg; c++){
                    memset(p->address,k,p->length);
                    p++;
                }
                sgat_dio_exit_req(&gfops,&dio_req,1);
            }
            return 0;
            break;
#ifdef CONFIG_TIVO_EIGER
            case SATA_UNPLUG_INT:
                /* This is a blocking call */
                {
                    int port, offset, regvalue;

                    port = 1; /* Currently hard coding to port 1 */

                    if(port >= NUM_SATA_PORTS)
                        return -EINVAL;

                    //printk("entering unplug %d \n", port);

                    /* SATA STATUS port */
                    offset = MMIO_SATA_STATUS + port*PORT_OFFSET;

                    /* Check the status */
                    regvalue = mmio_inl(offset);
                    regvalue &= PHY_IN_BIST;

                    if(regvalue)
                        return 0;

                    /* If user request for non blocking */
                    if(arg & 0x2)
                    {
                        /* Any positive number is treated error output of command (not ioctl error).
                        */
                        return 1;
                    }
                    //printk("Going to sleep for interrupt %d\n", __LINE__);

                    /* Now we have to wait for interrupt */
                    return sata_wait_for_IRQ(port, offset, PHY_IN_BIST);

                    break;
                }
            case SATA_PLUG_INT:
                /* This is a blocking call */
                {
                    int port, offset, regvalue;

                    port = 1; /* Currently hard coding to port 1 */

                    if(port >= NUM_SATA_PORTS)
                        return -EINVAL;

                    //printk("entering plug %d \n", port);

                    /* SATA STATUS port */
                    offset = MMIO_SATA_STATUS + port*PORT_OFFSET;

                    /* Check the status */
                    regvalue = mmio_inl(offset);
                    regvalue &= PHY_IS_RDY;

                    if(regvalue)
                        return 0;

                    /* If user request for non blocking */
                    if(arg & 0x2)
                    {
                        /* Any positive number is treated error output of command (not ioctl error).
							*/
							return 1;
						}
						/* Now we have to wait for interrupt */
						return sata_wait_for_IRQ(port, offset, PHY_IS_RDY);
						break;
					}
#endif
            case LOCKMEM:
                printk("gen_ioctl: LOCKMEM command STUBBED\n");
                return 0;
            break;
            case UNLOCKMEM:
                printk("gen_ioctl: UNLOCKMEM command STUBBED\n");
            return 0;
            break;

            case IS_PM_FWUPGD_REQD:
            {
                int ret = -EINVAL;
#if SKIP_SATA_PM_FW_UPGD
                return -EACCES;
#endif
#ifdef CONFIG_TIVO_DEVEL
                extern void sw_watchdog_threads_hold(void);
                extern void sw_watchdog_threads_release(void);

                /* Turn off sw watchdog for a while*/
                sw_watchdog_threads_hold();
#endif
                if ( is_fw_upgd_reqd)
                    ret=is_fw_upgd_reqd();
#ifdef CONFIG_TIVO_DEVEL
                sw_watchdog_threads_release();
#endif
                return ret;
            }
            break;

            case DWNLD_SATA_PM_FW:
            {
                int ret = -EINVAL;
#if SKIP_SATA_PM_FW_UPGD
                return -EACCES;
#endif
#ifdef CONFIG_TIVO_DEVEL
                extern void sw_watchdog_threads_hold(void);
                extern void sw_watchdog_threads_release(void);
                /* Turn off sw watchdog for a while until we complete
                 * upgrading FW in blocking PIO mode
                 */ 
                sw_watchdog_threads_hold();
#endif
                if ( upgd_fw_init )
                    ret=upgd_fw_init();
#ifdef CONFIG_TIVO_DEVEL
                /*Trun on sw wdog*/
                sw_watchdog_threads_release();
#endif
                return ret;
            }
            break;
            default:
                printk(KERN_ERR "gen_ioctl: Invalid command\n");
                return -EINVAL;
        }
       return -EINVAL; 
}

static struct file_operations gioctl_fops = {
        ioctl:   gioctl_ioctl,
        open:    gioctl_open,
};

void __init init_gioctl_if(void){
        int result;
        result = register_chrdev(GIOCTL_MAJOR, "gioctl", &gioctl_fops);
        if ( result < 0 )
                printk(KERN_ERR "gioctl: "
                                "failed to register char dev [major %d]\n",GIOCTL_MAJOR);
}

module_init(init_gioctl_if);
